import argparse
import os
import re
from pathlib import Path

import pytest
from carta_valet.valet import (
    connect_to_redis,
    duplicate_redis_values_found,
    generate_random_str,
    get_available_port,
    get_redis_values,
    run_carta,
    set_args,
)


@pytest.fixture
def redis():
    """
    Pytest fixture that provides a sandbox Redis server for testing purposes

    :return Sandbox Redis server
    """
    import fakeredis

    return fakeredis.FakeStrictRedis()


def test_generate_random_str():
    """
    Tests that generate_random_str returns a random sequence of 12 letters and digits
    """
    assert re.match(r"^[A-z0-9]{12}$", generate_random_str())


def test_get_available_port():
    """
    Tests that get_available_port returns an available port on localhost
    """
    import socket

    with socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM) as s:
        try:
            s.bind(("", get_available_port()))
        except OSError as e:
            # Port not available
            print(e)
            assert False


@pytest.mark.skip(reason="Connects to a real Redis server")
def test_connect_to_redis_success():
    """
    Tests that a connection can be made to the Redis server
    """
    args = set_args(make_default_args())
    connect_to_redis(args.redis_host, args.redis_port, args.redis_password)


def test_connect_to_redis_error():
    """
    Tests that a connection to an invalid Redis server correctly throws an exception
    """
    import redis

    with pytest.raises(redis.ConnectionError):
        connect_to_redis("localhost", 6666)


def test_get_redis_values():
    """
    Tests that filled out Redis values are given correctly
    """
    random_id = generate_random_str()
    front_end_port = 7777
    back_end_port = 6464
    reverse_proxy_host = "test"
    front_end = f"carta-front-{random_id}"
    back_end = f"carta-back-{random_id}"
    values = [
        f"traefik/http/routers/{front_end}/rule",
        f"traefik/http/routers/{front_end}/service",
        f"traefik/http/services/{front_end}/loadbalancer/servers/0/url",
        f"traefik/http/routers/{front_end}/middlewares/0",
        f"traefik/http/routers/{back_end}/rule",
        f"traefik/http/routers/{back_end}/service",
        f"traefik/http/services/{back_end}/loadbalancer/servers/0/url",
        f"traefik/http/routers/{back_end}/middlewares/0",
        f"traefik/http/routers/{back_end}/middlewares/1",
    ]
    assert (
        value
        in get_redis_values(
            reverse_proxy_host, random_id, random_id, front_end_port, back_end_port
        ).keys()
        for value in values
    )


def make_default_args() -> argparse.Namespace:
    """
    Create argparse namespace with default arguments that the valet will set

    :return: Namespace with default args set
    """
    os.environ["CAPO_PROFILE"] = "local"
    default_args = argparse.Namespace()
    default_args.profile = "local"
    default_args.timeout = 1
    default_args.where = "."
    return default_args


def test_set_args():
    """
    Tests that the set_args function gives predictable output back
    """
    # Testing default args
    mock_args = argparse.Namespace(timeout=None, where=None)
    args = set_args(mock_args)
    assert args.profile.profile == os.environ.get("CAPO_PROFILE")
    assert args.timeout == 1
    assert args.where == Path().cwd().absolute()


def test_duplicate_redis_values_found(redis):
    """
    Tests that duplicate Redis values are correctly reported

    :param redis: Custom fixture that provides a sandbox Redis server
    """
    redis.set("duplicate", "value")
    assert duplicate_redis_values_found({"duplicate": "value"}, redis) is True
    assert duplicate_redis_values_found({"not_duplicate": "value"}, redis) is False


def test_run_carta(mocker, caplog):
    """
    Test that CARTA runs successfully with setup and teardown

    :param mocker: Mocker fixture from pytest-mocker
    :param caplog: Built-in pytest fixture that allows viewing of logs
    """
    # Mock subprocess call so subprocess is not created
    mocker.patch("subprocess.Popen")
    run_carta("fake_carta_path", 1, Path(), 6464, 7777, "carta_url")
    messages = [record.message for record in caplog.records]
    assert "To access CARTA, enter this URL into a web browser: carta_url" in messages
